package MSN;

# A module to allow msn stuff to happen in perl :-)
# (C)opyright Mark Zealey <mark@zealos.org> 2002
use strict;
use Digest::MD5 qw/md5_hex/;
use Socket;
use IO::Handle;

# Use high-res timing if avaliable
BEGIN { eval 'use Time::HiRes qw/time/' }

# The usual msn ports
use constant DEFAULT_MSN_PORT => 1863;

# The standard header we use to send a message
our $MSN_STD_HDR = <<EOF;
MIME-Version: 1.0
Content-Type: text/plain; charset=UTF-8
X-MMS-IM-Format: FN=MS%20Sans%20Serif; EF=; CO=0; CS=0; PF=0

EOF

# The standard header for file transfer communication
our $MSN_FILE_HDR = <<EOF;
MIME-Version: 1.0
Content-Type: text/x-msmsgsinvite; charset=UTF-8

EOF

# The standard header for user is typing messages
our $MSN_TYPE_HDR = <<EOF;
MIME-Version: 1.0
Content-Type: text/x-msmsgscontrol
TypingUser: 
EOF

# Version string we send to server on connection
our $MSN_VERS = 'MSNP7 MSNP6 MSNP5 MSNP4 CVR0';

# The magic we add to a CHL response. Looks like M$ tried to kill us *sigh*
our $MSN_QRY_MAGIC = 'Q1P7W2E4J9R8U3S5';

# Hash of all the different commands we can send to it.
our %MSNCMD = (
		version => 'VER',
		ack => 'ACK',
		info => 'INF',
		change_servers => 'XFR',
		user => 'USR',
		sync => 'SYN',
		change_fuser => 'REA',
		change_my_state => 'CHG',
		quit => 'BYE',
		join => 'JOI',
		message => 'MSG',
		list => 'LST',
		swb_list => 'IRO',
		add => 'ADD',
		del => 'REM',
		ring => 'RNG',
		answer => 'ANS',
		initial_state => 'ILN',
		offline => 'FLN',
		change_state => 'NLN',
		invite => 'CAL',
		out => 'OUT',
		remove => 'REM',
		size => 'FIL',
		transfer => 'TFR',
		challange => 'CHL',
		query => 'QRY',
		ping => 'PNG',
		pong => 'QNG',		# Neat ;)
	);

# Hash to convert errors (as numbers) into human text
our %ERRORS = (
		200 => 'Syntax error',
		201 => 'Invalid parameter',
		205 => 'Invalid user',
		206 => 'Domain name missing',
		207 => 'Already logged in',
		208 => 'Invalid username',
		209 => 'Invalid fusername',
		210 => 'User list full',
		215 => 'User already there',
		216 => 'User already on list',
		217 => 'User not online',
		218 => 'Already in mode',
		219 => 'User is in the opposit list',
		280 => 'Switchboard failed',
		281 => 'Transfer to switchboard failed',

		300 => 'Required field missing',
		302 => 'Not logged in',

		500 => 'Internal server error',
		501 => 'Database server error',
		510 => 'File operation failed',
		520 => 'Memory allocation failed',

		600 => 'Server is busy',
		601 => 'Server is unavaliable',
		602 => 'Peer nameserver is down',
		603 => 'Database connection failed',
		604 => 'Server is going down',

		707 => 'Could not create connection',
		711 => 'Write is blocking',
		712 => 'Session is overloaded',
		713 => 'Too many active users',
		714 => 'Too many sessions',
		715 => 'Not expected',
		717 => 'Bad friend file',

		911 => 'Authentication failed',
		913 => 'Not allowed when offline',
		920 => 'Not accepting new users',
	);

# Hash to convert different states to messanger format
our %STATES = (
		online => 'NLN',
		offline => 'FLN',
		busy => 'BSY',
		idle => 'IDL',
		hidden => 'HDN',
		brb => 'BRB',
		phone => 'PHN',
		lunch => 'LUN',
		away => 'AWY',
	);

# Reasons for being kicked off the server
our %OUTS = (
		OTH => 'You have logged in from another computer',
		SSD => 'Server going down for an upgrade',
	);

our %REV_STATES = reverse %STATES;

# Select bits
our %SELECT = (
		r => 0,
		w => 1,
		e => 2,
	);

sub new {
	my $class = shift;
	$class = ref $class || $class;

	my $self = {
			@_,
			session => 0,	# Main connection ID
			rings => [],
			events => {},
			files => [],
			state => 'Not connected',
			buffer => '',
			deb => 'Main',
			sel => ['', '', ''],	# Select bits
			fd => [{}, {}, {}],	# fh hash
		};

	$self->{Debug} ||= undef;
	defined $self->{Debug} and $self->{Debug} == 1 and $self->{Debug} =
								\*STDERR;
	
	bless $self, $class;
	$self->{MSN} = $self;
}

sub add {
	my $self = shift;
	%$self = (%$self, @_)
}

sub connected {
	my $self = shift;

	# If we are disconnected, fileno() or getpeername() will return undef
	$self and $self->{sock} and fileno $self->{sock} and
						getpeername $self->{sock}
}

sub con {
	my $self = shift;

	$self->{sock} = new IO::Handle;
	socket($self->{sock}, PF_INET, SOCK_STREAM, getprotobyname 'tcp') or
				return $self->err("socket() returned $!");

	if(!defined $self->{sock}->blocking(0)) {
		undef $self->{sock};
		return $self->err("Error with setting blocking: $!")
	}

	my ($host, $port) = $self->{server} =~ /^([^:]+)(?::(\d+))?$/
					or return $self->err('Invalid server');

	if(!CORE::connect $self->{sock}, sockaddr_in($port, inet_aton $host)
					and $! ne 'Operation now in progress') {
		undef $self->{sock};
		return $self->err("connect() returned $!")
	}

	$self->{sock}->autoflush(1);

	1
}

sub connect {
	my $self = shift;

	$self->add(@_);

	$self->{state} = 'connecting';
	$self->{trans} = 1; 

	$self->{Host} ||= 'messenger.hotmail.com';
	$self->{Port} ||= DEFAULT_MSN_PORT;

	return $self->err("No username defined") unless $self->{User};
	return $self->err("No password defined") unless $self->{Pass};

	# Fixup the username for a lazy luser
	$self->{User} = $self->fix_user($self->{User});
	$self->{FUser} ||= "I'm too lame to change my fusername";

	$self->{server} = $self->{Server} || "$self->{Host}:$self->{Port}";
	$self->{users} = {};

	$self->debug("Connecting to $self->{server}");

	return $self->err unless $self->con;
	$self->{sock_fn} = fileno $self->{sock};

	$self->add_fh('r', $self->{sock_fn}, $self);
	$self->add_fh('w', $self->{sock_fn}, $self);

	1
}

# Change your fuser-name aka 'screen name'
sub change_fuser {
	my ($self, $name) = @_;

	$self = $self->{MSN};
	$name ||= $self->{FUser};
	$name =~ s/([^\w\d])/sprintf '%%%2x', ord $1/eg;

	$self->send_line('change_fuser', $self->{User}, $name)
}

sub change_my_state {
	my ($self, $state) = @_;

	exists $STATES{lc $state} or return undef;

	# Already in this state? Don't waste bw :-)
	exists $self->{state} and lc $self->{state} eq lc $state and return 1;

	$self->send_line('change_my_state', $STATES{lc $state});
}

sub get_state { shift->{MSN}{state} }

sub disconnect {
	my $self = shift;

	$self->{state} = 'Not connected';

	$self->del_fh('r', $self->{sock_fn});
	$self->del_fh('w', $self->{sock_fn});

	# Kill all our ring connections
	$_->disconnect for grep {defined} @{$self->{rings}};

	if($self->connected) {
		$self->send_line('out');
		close $self->{sock}
	}

	delete $self->{connected}
}

sub add_fh {
	my ($self, $h, $fileno, $obj) = @_;

	$h = $SELECT{$h};

	# If we were passed a fileno, don't fileno it :)
	$fileno = fileno $fileno if ref $fileno;

	vec($self->{MSN}{sel}[$h], $fileno, 1) = 1;
	$self->{MSN}{fd}[$h]{$fileno} = $obj
}

sub del_fh {
	my ($self, $h, $fileno) = @_;

	$h = $SELECT{$h};

	# If we were passed a fileno, don't fileno it :)
	$fileno = fileno $fileno if ref $fileno;

	vec($self->{MSN}{sel}[$h], $fileno, 1) = 0;
	delete $self->{MSN}{fd}[$h]{$fileno}
}

# See if there's any stuff we need to handle and handle it.
#
# Optionally pass a variable saying if you want timeout to it and then 3
# bit-patterns for read, write and error polling respectivly.
#
# Returns an array ref of 3 bit patterns, in select() style.
sub do_stuff {
	my ($self, $loop, $timeout) = (shift, shift, shift);

	# Build a list of our user-specified fh's
	my @user;
	$user[$_] = $_[$_] || '' for values %SELECT;

	# Anything to do?
	do {{
		# Not connected, nothing to do..
		if(!$self->connected and $self->{connected}) {
			return $self->call_callback('disconnected')
		}

		# Make our bitmap
		my @s;
		$s[$_] = $self->{sel}[$_] | $user[$_] for values %SELECT;

		# Select all the bits
		last if 0 == select($s[0], $s[1], $s[2], $timeout);

		my (%intern, $ret);
		for my $n (0 .. $#s) {
			# If the thing is not ours.. it must be a lusers!
			for(keys %{$self->{fd}[$n]}) {
				next if 0 == vec $s[$n], $_, 1;

				vec($s[$n], $_, 1) = 0;
				$intern{$_} = $self->{fd}[$n]{$_}
			}

			$ret = 1 if !$ret and $s[$n] ne (chr(0) x length $s[$n])
		}

		return [@s] if $ret;

		# Handle ours now
		for(values %intern) {
			next unless $_;

			$_->internal_do_stuff or ($_ and $_->disconnect)
		}
	}} while($loop);

	[]
}

sub internal_do_stuff {{
	my $self = shift;

	return 1 unless $self->connected;

	if(!$self->{connected}) {
		$self->{connected} = 1;		# Prevent lots of these...

		if(!$self->get_sess_id) {		# Main one
			# Send version
			$self->send_line('version', $MSN_VERS) or
							return $self->err
		} else {				# Switchboard
			# We opened successfully, call the open_chat callback
			$self->call_callback('open_chat', $self->get_sess_id);

			if(exists $self->{ID}) {    # We've been invited here
				# We're now connected, auth ourselves
				$self->send_line('answer', @{%$self}{'Me',
							qw/Auth ID/}) or
							return $self->err
			} else {			# We set this chat up
				# Auth ourselves
				$self->send_line('user', @{%$self}{qw/Me Auth/})
							or return $self->err;
			}
		}

		$self->call_del_event('connect', $self);
		$self->del_fh('w', $self->{sock_fn});
		return 1
	}

	$self->get_block or return $self->err;

	if($self->{msg_wait}) {			# Waiting for a message
		return 1 if length $self->{buffer} < $self->{msg_wait}[0];

		# Got a full message, handle it in handle_msg()
		my $msg = substr $self->{buffer}, 0, $self->{msg_wait}[0], '';
		$self->handle_msg(@{$self->{msg_wait}}[1,2], $msg);

		undef $self->{msg_wait}
	}

	# o_O we have something to do!
	my $line = $self->get_line or return 1;

	my @bit = split ' ', $line;

	# First check if this event is one someone else wants to catch
	if($bit[1] and $bit[1] =~ /^\d+$/ and exists $self->{events}{$bit[1]}) {
		return $self->call_del_event($bit[1], $self, $line)
	}

	# Else, look through our handlers
	if($bit[0] eq $MSNCMD{change_servers}) {
		return $self->change_server(Server => $bit[3])
	} elsif($bit[0] eq $MSNCMD{message}) {	# We got a message
		# Message format, USER LENGTH where LENGTH is the length
		# to read in bytes, neat
		$self->{msg_wait} = [@bit[3,1,2]]
	} elsif($bit[0] eq $MSNCMD{list}) {	# List maintainance
		$self->add_user(@bit[2,6,7]) if defined $bit[7]
	} elsif($bit[0] eq $MSNCMD{add}) {
		# Index here == 0...
		my $u = $self->add_user(@bit[2,4,5]);

		if($bit[2] eq 'AL') {
			$self->call_callback('auth_success', $u)
		} elsif($bit[2] eq 'BL') {
			$self->call_callback('block_success', $u)
		} elsif($bit[2] eq 'FL') {
			$self->call_callback('add_success', $u)
		}
	} elsif($bit[0] eq $MSNCMD{del}) {
		my $u = $self->del_user(@bit[4,2]);

		if($bit[2] eq 'AL') {
			$self->call_callback('deauth_success', $u);
		} elsif($bit[2] eq 'BL') {
			$self->call_callback('unblock_success', $u);
		} elsif($bit[2] eq 'FL') {
			$self->call_callback('del_success', $u);
		} elsif($bit[2] eq 'RL') {
			$self->call_callback('del', $u);
		}
	} elsif($bit[0] eq $MSNCMD{ring}) {	# Someone is ringing us
		my $swb = MSN::Switchboard->new(
						ID => $bit[1],
						Server => $bit[2],
						Auth => $bit[4],
						MSN => $self,
					) or return;

		$self->{MSN}{rings}[$swb->get_sess_id] = $swb
	} elsif($bit[0] eq $MSNCMD{join}) {
		# Add the user
		$self->_join_user(@bit[1,2]);
	} elsif($bit[0] eq $MSNCMD{quit}) {
		$self->leave($bit[1], $self->get_sess_id);
	} elsif($bit[0] eq $MSNCMD{out}) {
		$self->call_callback('out', $OUTS{$bit[1]});
		$self->disconnect;
		return $self->err("OUT recieved: $bit[1]")
	} elsif($bit[0] eq $MSNCMD{change_my_state}) {
		$self->{state} = $REV_STATES{$bit[2]};
		$self->call_callback('state_success', $bit[2])
	} elsif($bit[0] eq $MSNCMD{change_fuser}) {
		return $self->err unless $bit[3] eq $self->{User};

		$self->{FUser} = $bit[4];

		$self->call_callback('fuser_success', $self->{FUser});
	} elsif($bit[0] eq $MSNCMD{challange}) {
		$self->send_line('query', 'msmsgs@msnmsgr.com 32') or
							return $self->err;
		print {$self->{sock}} md5_hex($bit[2] . $MSN_QRY_MAGIC);
	} elsif($bit[0] eq $MSNCMD{pong}) {
		my $sent = shift @{$self->{pings}};
		$sent or return $self->err;

		$self->call_callback('ping', time - $sent)

	# With these three, we just change the user's state, dunno why the
	# protocol didn't originally do this...
	} elsif($bit[0] eq $MSNCMD{initial_state}) {
		$self->change_user_state(@bit[2..4])
	} elsif($bit[0] eq $MSNCMD{change_state}) {
		$self->change_user_state(@bit[1..3])
	} elsif($bit[0] eq $MSNCMD{offline}) {
		# First, disconnect the user from the swb sessions. (Some buggy
		# clients don't BYE from the swb sessions when they quit...)
		my $u = $self->get_user($bit[1]);

		$self->leave($bit[1], $_) for grep /^\d+$/, keys %{$u->{Lists}};

		# And change the users state to offline
		$self->change_user_state($STATES{offline}, $bit[1])

	# Error handling
	} elsif($bit[0] =~ /^(\d{3})$/) {		# Usially means an error
		my $err;
		$err = exists $ERRORS{$1} ? $ERRORS{$1} : "Unknown error $line";
		$self->err("Error: $err");
		$self->call_callback('error', $err)

	# When joining a switchboard:
	} elsif($bit[0] eq $MSNCMD{swb_list}) {
		$self->_join_user(@bit[4,5]);
	} elsif($bit[0] eq $MSNCMD{answer}) {
		# Ignore it.

	# These next ones are only used in the initalization process:
	} elsif($bit[0] eq $MSNCMD{version}) {
		if($bit[2] eq '0') {
			$self->disconnect;

			return $self->err('Protocol not supported')
		}

		$self->send_line('info', '') or return $self->err;

	# Server responds with the auth protocol we use
	} elsif($bit[0] eq $MSNCMD{info}) {
		return $self->err("Auth protocol $bit[2] not supported")
							if $bit[2] ne 'MD5';

		# ACK it and send username
		$self->send_line('user', "MD5 I $self->{User}")
							or return $self->err;
	} elsif($bit[0] eq $MSNCMD{user}) {
		if($bit[2] eq 'MD5') {		# Send pword
			$self->send_line('user', 'MD5 S',
					md5_hex("$bit[4]$self->{Pass}"))
							or return $self->err
		} elsif($bit[2] eq 'OK') {	# Whee, auth'd properly
			return 1 if $self->get_sess_id;		# Don't on swb's
			$self->change_fuser or return $self->err;
			$self->sync or return $self->err;

			# Announce to the world that we're around!
			$self->change_my_state('online') or return $self->err;

			$self->call_callback('connect')
		} else {
			$self->debug("Not sure about '$line'")
		}
	} else {
		$self->debug("Not sure about '$line'");
	}

	push @_, $self;				# XXX: KLUGE!! :(
	redo
}}

# Sync says what the status of the local lists are.. as we don't have local
# lists, we just send 1 so it gives us everything =)
sub sync { shift->send_line('sync', 1) }

sub ping {
	my $self = shift;
	$self->send_line('ping') or return $self->err;
	push @{$self->{pings}}, time;
}

# A user has left a switchboard session
sub leave {
	my ($self, $user, $sess) = @_;

	my $u = $self->del_user($user, $sess);

	my $swb = $self->{MSN}{rings}[$sess];

	return unless defined $swb;

	delete $swb->{User}{$user};

	$swb->call_callback('part', $u);

	# Disconnect unless we have more people here.
	$swb->disconnect unless keys %{$swb->{User}}
}

# Change the state of a user.
sub change_user_state {
	my ($self, $state, $user, $fuser) = @_;

	my $u = $self->get_user($user, $fuser) or return
						$self->unknown_user($user);

	# Return if the states are equal
	return 1 if $u->{State} eq $state;

	$u->{State} = $state;

	$self->call_callback('change_state', $u)
}

sub start_ring { shift->_start_ring(undef, @_) }

sub _start_ring {
	my $self = shift;

	# Get info about a new switchboard to connect to
	my $id = $self->send_line('change_servers', 'SB') or return $self->err;

	$self->add_event($id, \&_ring_response, @_)
}

sub _ring_response {
	my ($self, $line, $ftp, @users) = @_;

	my @bit = split ' ', $line;

	my $swb = MSN::Switchboard->new(
					Server => $bit[3],
					Auth => $bit[5],
					MSN => $self,
				) or return $self->err;

	# Invite the users
	$swb->add_event('connect', sub { shift->invite_users(undef, @_) },
								@users);
	$self->{MSN}{rings}[$swb->get_sess_id] = $swb;
}

sub invite_users {
	my ($self, $swb_num) = (shift, shift);

	my $swb = $self;
	$swb = $self->{MSN}{rings}[$swb_num] if $swb_num;

	return $self->err unless defined $swb;

	# Invite the users to the switchboard now
	$swb->_invite_user($_) for @_;
}

sub state {
	my ($self, $state) = @_;

	return $REV_STATES{$state} if exists $REV_STATES{$state};
	'Unknown'
}

# Return the state in a human-readable format
sub get_human_state {
	my ($self, $user) = @_;
	my $u = $self->get_user($user) or return $self->unknown_user($user);
	$self->state($u->{State})
}

sub handle_msg {
	my ($self, $user, $fuser, $msg) = @_;
	my %message;

	# No fuser needed, msn keeps the same fuser when talking to you even if
	# the luser has changed it. Silly Billy...
	$self->get_user($user);

	$message{User} = $user;

	($message{Raw} = $msg) =~ s/\r//g;

	# With any luck, we got here with no errors.
	($message{MIMEVer}) = $message{Raw} =~ /MIME-Version: (.*)\n/i;

	$self->debug("Mime version $message{MIMEVer} not supported fully")
				unless $message{MIMEVer} == 1;

	($message{Type}) = $message{Raw} =~ /Content-Type: (.*?)[;\n]/i;

	if($message{Type} eq 'text/plain') {
		# A normal message
		($message{Body}) = $message{Raw} =~ /\n\n(.*?)$/is;
		$message{Body} ||= '';
		return $self->call_callback('message', \%message);
	} elsif($message{Type} eq 'application/x-msmsgsemailnotification') {
		# An email message - Added by reggy
		my $t = 0;
		($message{From}) = $message{Raw} =~ /From: (.*)/i or $t = 1;
		($message{Subject}) = $message{Raw} =~ /Subject: (.*)/i
								or $t = 1;
		($message{FromAddr}) = $message{Raw} =~ /From-Addr: (.*)/i
								or $t = 1;

		return $self->call_callback('new_mail', \%message) if $t == 0;
	} elsif($message{Type} eq 'text/x-msmsgscontrol') {
		return $self->call_callback('user_types', \%message)
	} elsif($message{Type} eq 'text/x-msmsgsinvite') {	# File transfer
		for(split /\n/, $message{Raw}) {
			/^([^:]+):\s*(.*)$/ and $message{$1} = $2
		}

		if(exists $message{'Application-Name'} and
			$message{'Application-Name'} eq 'voice conversation') {
			$self->debug('Voice not supported. (yet :)');
			$self->debug($message{Raw});
			return
		}

		# Abreviate these..
		my %t = (	'Application-FileSize' => 'Size',
				'Application-File' => 'File',
				'IP-Address' => 'Host',
				'Invitation-Cookie' => 'Cookie',
				'AuthCookie' => 'Auth',
			);

		for(keys %t) {
			$message{$t{$_}} = $message{$_} if exists $message{$_}
		}

		# XXX debug voice stuff.
		$self->debug($message{Raw}) unless exists $message{File};
		
		$message{Sess} = $self->get_sess_id;

		if($message{'Invitation-Command'} eq 'INVITE') {
			my $ftp = MSN::Recieve->new (
							MSN => $self->{MSN},
							SWB => $self,
							%message
						) or return $self->err;

			return $self->{MSN}->call_callback('file_invite', $ftp);
		} elsif($message{'Invitation-Command'} eq 'ACCEPT') {
			my $ftp = $self->{MSN}{_FTPCookie}{$message{Cookie}};
			return $self->err unless $ftp;

			# If we are recieving this file
			if(exists $message{Port}) {
				$ftp->add(%message);

				$self->{MSN}->call_callback('file_start', $ftp);

				# We can start downloading now!
				return $ftp->connect
			} else {		# We're sending
				my $auth = int rand 2**32;
				# Fireup our ftp session
				$ftp->connect(Auth => $auth);

				return $self->send_file_msg(
					'Invitation-Cookie' => $message{Cookie},
					'Invitation-Command' => 'ACCEPT',
					'IP-Address' => $self->{MSN}{IP},
					'Port' => $ftp->{Port},
					'AuthCookie' => $auth
				)
			}
		}
	}

	$self->debug("Message type $message{Type} not supported");
	$self->debug($message{Raw});
}

sub file_get {
	my ($self, $num, $get) = @_;

	my $message = $self->{MSN}{FTPs}[$num] or return $self->err;

	$message->{SWB}->send_file_msg(
			'Invitation-Command' => $get ? 'ACCEPT' : 'REJECT',
			'Invitation-Cookie' => $message->{Cookie}
		)
}

sub file_send {
	my $self = shift;
	my $swb = $self->{MSN}{rings}[$_[0]];
	return $self->err unless $swb;
	$swb->_file_send(@_)
}

sub _file_send {
	my ($self, $user, $file, $size) = @_;

	# Find out our IP address
	if(!$self->{MSN}{IP}) {
		my $sock = getsockname $self->{sock};
		$self->{MSN}{IP} = inet_ntoa +(sockaddr_in $sock)[1]
	}

	return $self->err("File $file does not exist") unless -f $file;

	$size ||= -s $file;
	return $self->err("Size $size bytes is invalid") if !$size or $size < 1;

	my ($nice_file) = $file =~ m!([^/]+)$!;

	my %msg = (
			'Invitation-Cookie' => int rand 2**32,
			'Invitation-Command' => 'INVITE',
			'Application-Name' => 'Fooie',
			'Application-GUID' => '{5D3E02AB-6190-11d3-BBBB-00C04F795683}',
			'Application-FileSize' => $size,
			'Application-File' => $nice_file,
		);

	my $fh;
	open $fh, '<', $file or return $self->err("Can't open $file, $!");

	MSN::Send->new(
				MSN => $self->{MSN},
				SWB => $self,
				Size => $size,
				FH => $fh,
				Cookie => $msg{'Invitation-Cookie'},
			);

	$self->send_file_msg(%msg)
}

sub send_msg_swb {
	my ($self, $swbnum) = (shift, shift);
	return undef unless $self->{MSN}{rings}[$swbnum];

	$self->{MSN}{rings}[$swbnum]->send_msg(@_)
}

sub send_typing_swb {
	my ($self, $swbnum) = (shift, shift);
	return undef unless $self->{MSN}{rings}[$swbnum];

	$self->{MSN}{rings}[$swbnum]->send_typing_msg
}

sub send_msg { shift->send_std_msg($MSN_STD_HDR, @_) }

sub send_typing_msg {
	my $self = shift;
	$self->send_std_msg($MSN_TYPE_HDR . $self->{MSN}{User})
}

sub send_file_msg {
	my ($self, %opts, $str) = @_;

	$str .= "$_: $opts{$_}\n" for keys %opts;
	$self->debug($str);

	$self->send_std_msg($MSN_FILE_HDR, $str)
}

sub send_std_msg {
	my ($self, $msg) = (shift, shift);

	$msg .= join ' ', @_;

	$msg =~ s/\r?\n/\r\n/g;			# Update line-endings

	$self->send_line('message', 'N', length $msg) or return $self->err;

	print {$self->{sock}} $msg
}

# Add an event that will callback the specified function
sub add_event {
	my ($self, $id, $callback, @args) = @_;

	$self->{events}{$id} = [$callback, [@args]];
}

# Remove event, call and return
sub call_del_event {
	my ($self, $id, @args) = @_;

	my $t = $self->{events}{$id};

	&{$t->[0]}(@args, @{$t->[1]}) if $t;
	delete $self->{events}{$id};

	1
}

sub sess_exists {
	my ($self, $num) = @_;
	$num =~ /\D/ and return undef;
	defined $self->{MSN}{rings}[$num]
}

sub part_sess {
	my ($self, $num) = (shift, shift);
	return $self->err("Session $num non-existant")
						unless $self->sess_exists($num);

	$self->{MSN}{rings}[$num]->disconnect(@_);

	1
}

# Get a unique session ID
sub get_usid {
	my $self = shift;

	my $i = 1;

	$i++ while $self->{MSN}{rings}[$i];

	$i
}

sub get_sess_id { shift->{session} }

sub add_callback {
	my ($self, $name, $func) = @_;
	push @{$self->{MSN}{callbacks}{$name}}, $func;
}

sub del_callback {
	my ($self, $name, $func) = @_;

	if(defined $func) {
		my $c = $self->{MSN}{callbacks}{$name};

		@$c = grep $_ ne $func, @$c
	} else {
		delete $self->{MSN}{callbacks}{$name}
	}
}

sub call_callback {
	my ($self, $name) = (shift, shift);

	for my $t (@{$self->{MSN}{callbacks}{$name}}) {
		eval { &{$t}($self, @_) };
		$self->debug("Hmm, function didn't work, $@") if $@
	}
}

# AL -> Allow List (Ones who are friends with me)
# BL -> Block List (My enemys)
# FL -> Friends List (Ones I'm friends with)
# RL -> Request List (Ones that want to be my friend)

sub _user_is_in_list {
	my ($self, $list, $user, $fuser) = @_;
	my $u = $self->get_user($user, $fuser) or return
						$self->unknown_user($user);

	exists $u->{Lists}{$list} 
}

sub user_is_allowed { shift->_user_is_in_list('AL', @_) }
sub user_is_blocked { shift->_user_is_in_list('BL', @_) }
sub user_is_friend { shift->_user_is_in_list('FL', @_) }
sub user_is_request { shift->_user_is_in_list('RL', @_) }

# Fix a username
sub fix_user {
	my ($self, $user) = @_;

	$user =~ /^([^@]+)(.*)/;

	lc $1 . ($2 || '@hotmail.com')
}

# Return a plain-text decoded fuser
sub plain_fuser {
	my ($self, $fuser) = @_;
	$fuser =~ s/\%([0-9A-Fa-f]{2})/chr hex $1/eg;
	$fuser
}

# list name, name, friendly name, (optional) initial state
sub add_user {
	my ($self, $list, $user, $fuser, $state) = @_;

	$user = $self->fix_user($user);

	if(exists $self->{MSN}{users}{$user}) {		# Add list in
		$self->{MSN}{users}{$user}{Lists}{$list} = 1
	} else {					# Need to create new one
		($fuser) = $user =~ /^([^@]+)/ if !defined $fuser;

		$self->{MSN}{users}{$user} = {
					User => $user,
					FUser => $fuser,
					State => $state || $STATES{offline},
					Lists => {
						$list => 1
					},
				};
	}

	# If someone wants to be added, isn't already there and
	# isn't in our banned list, call the callback
	if($list eq 'RL' and !$self->user_is_allowed($user)
			and !$self->user_is_blocked($user)) {
		$self->call_callback('add', $self->get_user($user));
	}

	$self->{MSN}{users}{$user}
}

sub get_users {
	my $self = shift;

	%{$self->{MSN}{users}}
}

sub del_user {
	my ($self, $user, $list) = @_;

	my $u = $self->get_user($user) or return $self->unknown_user($user);

	if(defined $list) {
		$self->debug("Removing list $list from user $user");
		delete $u->{Lists}{$list};
	} else {
		# No list specified
		delete $self->{MSN}{users}{$u->{User}}
	}

	$u
}

# Same here for fuser as in add_user.
sub get_user {
	my ($self, $user, $fuser) = @_;

	$user = $self->fix_user($user);

	if(defined $fuser and $self->{MSN}{users}{$user}{FUser} ne $fuser) {
		my $old_fuser = $self->{MSN}{users}{$user}{FUser};
		$self->{MSN}{users}{$user}{FUser} = $fuser;
		$self->{MSN}->call_callback('fuser_change',
					$self->get_user($user), $old_fuser)
	}

	$self->{MSN}{users}{$user}
}

sub unknown_user {
	my ($self, $user) = @_;
	$self->call_callback('unknown_user', $user);
	1
}

sub do_contact {
	my ($self, $list, $user, $fuser) = @_;

	my $u = $self->get_user($user, $fuser) || $self->add_user($list,
								$user, $fuser);

	$self->send_line('add', "$list $u->{User} $u->{FUser}") or
							return $self->err;
	$self->add_user($list, $user);

	$u
}

sub rem_contact {
	my ($self, $list, $user) = @_;

	my $u = $self->get_user($user) or return $self->unknown_user($user);

	$self->send_line('del', $list, $u->{User}) or return $self->err;
	$self->del_user($user, $list);
}

sub authorize_contact	{ shift->do_contact('AL', @_) }
sub deauthorize_contact	{ shift->rem_contact('AL', @_) }
sub add_contact		{ shift->do_contact('FL', @_) }
sub del_contact		{ shift->rem_contact('FL', @_) }

sub block_contact { 
	my $self = shift;
	$self->rem_contact('AL', @_);
	$self->do_contact('BL', @_);
}

sub unblock_contact { 
	my $self = shift;
	$self->rem_contact('BL', @_);
	$self->do_contact('AL', @_);
}

# Disconnects and reconnects to the server in Host and Port.
sub change_server {
	my $self = shift;

	$self->disconnect;

	$self->connect(@_)
}

# Sends a line to the server, returns undef on failure, the contents of the line
# it sent otherwise.
sub send_line {
	my ($self, $str, @text) = @_;

	return $self->err("$str doesn't exist!") unless $MSNCMD{$str};

	my $send = $MSNCMD{$str};

	# Don't send a command number for file transfer
	if(@text) {
		$send .= ' ' . ++$self->{trans} if $self->{trans};
		$send .= " @text"
	}

	$self->debug("> $send");

	if(!print {$self->{sock}} "$send\r\n") {	# Disconnected
		$self->disconnect;
		return $self->err
	}

	$self->{trans} or 1
}

# Reads anything that can be read into the buffer. Returns undef on error.
sub get_block {
	my $self = shift;

	return $self->err('Not connected to server') unless $self->connected;

	my $t;
	my $ret = read $self->{sock}, $t, 2**20;	# Some large value..

	if(!defined $ret and $! !~ /temporarily unav/i) {	# Eweee.. error!
		$self->debug("Remote read error on session $self->{deb}: $!");

		$self->disconnect;

		return undef
	}

	# shove the rest into the buffer and see if we can read a line from it
	$self->{buffer} .= $t if $ret;

	1
}

# Returns the string got or undef on error/not a full string
sub get_line {
	my $self = shift;

	$self->{buffer} =~ s/(.*?)\r?\n//;
	$self->debug("< $1") if $1;

	$1
}

sub debug {
	my ($self, $t) = (shift, join(' ', @_));
	print {$self->{Debug}} "$self->{deb} $t\n" if $self->{Debug};
}

sub err {
	my ($self, $text) = (shift, join(' ', @_));

	if($text eq '') {
		my ($package, $file, $line) = caller;
		$text = "Unknown error at line $line, file $file"
	}

	$self->{Error} = $text;

	$self->debug("Error: $text");
	undef
}


# A separate Switchboard module is fired up for each chat session instance we
# wish to have. Some are also done for file transfer negociation.
package MSN::Switchboard;

our @ISA = qw(MSN);

sub new {
	my $class = shift;
	$class = ref $class || $class;

	my $self = {
			@_,
			buffer => '',
		};

	bless $self, $class;

	$self->{session} = $self->{MSN}->get_usid;
	$self->{trans} = 1; 
	$self->{Debug} ||= $self->{MSN}{Debug};
	$self->{Me} ||= $self->{MSN}{User};
	$self->{deb} = "SES$self->{session}";
	$self->{server} = $self->{Server} || "$self->{Host}:$self->{Port}";

	$self->debug("Connecting switchboard to $self->{server}");

	return $self->err unless $self->con;
	$self->{sock_fn} = fileno $self->{sock};

	$self->add_fh('r', $self->{sock_fn}, $self);
	$self->add_fh('w', $self->{sock_fn}, $self);

	$self
}

sub disconnect {
	my $self = shift;

	# Remove users that might have been involved in the chat
	$self->del_user($_, $self->get_sess_id) for keys %{$self->{User}};

	if($self->connected) {
		$self->send_line('out');
		$self->call_callback('close_chat', $self->get_sess_id);
		close $self->{sock};
		undef $self->{sock}
	}

	$self->del_fh('r', $self->{sock_fn});
	$self->del_fh('w', $self->{sock_fn});

	undef $self->{MSN}{rings}[$self->get_sess_id]
}

sub _invite_user {
	my ($self, $user) = @_;

	my $u = $self->get_user($user) or return $self->unknown_user($user);

	$self->send_line('invite', $u->{User})
}

sub _join_user {
	my ($self, $user, $fuser) = @_;

	# Update fname etc
	my $u = $self->add_user($self->get_sess_id, $user, $fuser);
	$self->{User}{$u->{User}} = 1;

	$self->call_callback('join', $u)
}

# Internal common bits for both MSN::Recieve and MSN::Send
package MSN::File;

our @ISA = qw(MSN);

our $MSN_FTP_VER = 'MSNFTP';

# Some magic number that the client sends to the host to say that it has
# successfully got the file.. hmm..
our $MSN_FTP_SUCCESS = 16777989;

sub new {
	my $class = shift;
	$class = ref $class || $class;

	my $self = {
			State => 'Waiting',
			@_,
			buffer => '',
		};

	bless $self, $class;

	return $self->err unless exists $self->{MSN};

	1 while $self->{MSN}{FTPs}[++$self->{Num}];

	$self->{MSN}{FTPs}[$self->{Num}] = $self;
	$self->{MSN}{_FTPCookie}{$self->{Cookie}} = $self;

	$self->{deb} = "FTP$self->{Num}";

	$self->{Debug} ||= $self->{MSN}{Debug};

	$self
}


# The module we use to recieve a file from the remote client
package MSN::Recieve;

our @ISA = qw(MSN::File);

sub connect {
	my $self = shift;
	$self->add(@_);

	$self->{State} = 'Connecting';
	$self->{downloading} = 0;		# Not downloading yet

	open $self->{FH}, '>', $self->{File} or return
			$self->err("Can't open $self->{File} for writing, $!");

	$self->{server} = $self->{Server} || "$self->{Host}:$self->{Port}";

	$self->con or return $self->err("Couldn't connect");
	$self->{sock_fn} = fileno $self->{sock};
	$self->debug('connecting');

	$self->add_fh('r', $self->{sock_fn}, $self);
	$self->add_fh('w', $self->{sock_fn}, $self);

	1
}

sub disconnect {
	my $self = shift;

	$self->{SWB}->call_callback('file_get_end', $self);

	delete $self->{MSN}{FTPs}[$self->{Num}];
	delete $self->{MSN}{_FTPCookie}{$self->{Cookie}};

	$self->del_fh('r', $self->{sock_fn});
	$self->del_fh('w', $self->{sock_fn});

	close $self->{sock} if defined $self->{sock};
	undef $self->{sock}
}

sub internal_do_stuff {
	my $self = shift;

	return undef if $self->{disconnect};		# XFR finished

	if(!$self->{connected}) {
		$self->{connected}++;
		$self->send_line('version', $MSN_FTP_VER) or return $self->err;

		# Don't poll for the ability to write now..
		return $self->del_fh('w', $self->{sock_fn})
	}

	$self->get_block or return $self->err;

	# If you have a hella-fast inet connection, I envy you, but, your client
	# will appear to lock while we download the file. This should stop the
	# client from appearing to lock up totally while downloading a file at
	# high speed.
	my $MAX_REPS = 50;

	if($self->{downloading}) {{	# Get a block, write to file
		return 1 unless $MAX_REPS--;

		# See if we got anything since last time
		$self->get_block;

		# 3 bytes is header:
		# Byte number | Meaning
		# 0           | unknown
		# 1           | Number of bytes in this chunk (lower byte)
		# 2           | Number of bytes in this chunk (higher byte)

		if(!$self->{next_chunk}) {
			return 1 unless length $self->{buffer} >= 3;
			my $t = substr $self->{buffer}, 0, 3, '';
			my @t = map {ord} split //, $t;
			$self->{next_chunk} = $t[2] * 256 + $t[1];
			$self->debug(@t, $self->{next_chunk})
		}

		return 1 unless length $self->{buffer} >= $self->{next_chunk};

		my $in = substr $self->{buffer}, 0, $self->{next_chunk}, '';

		print {$self->{FH}} $in;

		$self->{to_read} -= length $in;
		$self->debug("To read: $self->{to_read}");

		if(!$self->{to_read}) {			# Nothing left to read
			close $self->{FH};

			$self->send_line('quit', $MSN_FTP_SUCCESS);
			return $self->{disconnect} = 1;
		}

		undef $self->{next_chunk};

		redo
	}}

	while(defined(my $line = $self->get_line)) {
		my @bit = split ' ', $line;

		if($bit[0] eq $MSNCMD{version}) {
			return $self->err('Incompatable protocols:', $line)
						if $line !~ /$MSN_FTP_VER/;

			$self->send_line('user', $self->{MSN}{User},
					$self->{Auth}) or return $self->err;
		} elsif($bit[0] eq $MSNCMD{size}) {
			$self->{Size} = $bit[1];
			$self->{to_read} = $self->{Size};
			$self->{downloading} = 1;
			$self->send_line('transfer') or return $self->err
		}
	}

	1
}


# The module we use to recieve a file from the remote client
package MSN::Send;

our @ISA = qw(MSN::File);

use Socket;
use constant DEFAULT_FTP_PORT => 6891;
use constant MAX_XFER_SIZE => 2045;

sub connect {
	my $self = shift;

	$self->add(@_);

	$self->{serv} = new IO::Handle;
	socket($self->{serv}, PF_INET, SOCK_STREAM, getprotobyname 'tcp') or
				return $self->err("socket() returned $!");

	if(!defined $self->{serv}->blocking(0)) {
		undef $self->{serv};
		return $self->err("Error with setting blocking: $!")
	}

	# find a port we can use
	$self->{port} = 0;
	1 while defined $self->{MSN}{ftp_ports}[$self->{port}++];
	$self->{MSN}{ftp_ports}[$self->{port}] = 1;

	$self->{Port} = DEFAULT_FTP_PORT + $self->{port};

	bind($self->{serv}, sockaddr_in($self->{Port}, INADDR_ANY))
				or return $self->err("bind() returned $!");

	listen $self->{serv}, 1;

	$self->{serv_fn} = fileno $self->{serv};

	$self->add_fh('r', $self->{serv_fn}, $self);
	$self->add_fh('w', $self->{serv_fn}, $self);

	1
}

sub disconnect {
	my $self = shift;

	$self->{SWB}->call_callback('file_send_end', $self);

	# Free the port that we were using
	undef $self->{MSN}{ftp_ports}[$self->{port}];

	$self->disconnect_server;
	$self->disconnect_client
}

sub disconnect_server {
	my $self = shift;

	if(defined $self->{serv}) {
		close $self->{serv};
		undef $self->{serv};
		$self->del_fh('r', $self->{serv_fn});
		$self->del_fh('w', $self->{serv_fn})
	}
}

sub disconnect_client {
	my $self = shift;

	if(defined $self->{sock}) {
		close $self->{sock};
		undef $self->{sock};
		$self->del_fh('r', $self->{sock_fn});
		$self->del_fh('w', $self->{sock_fn})
	}
}

sub internal_do_stuff {{
	my $self = shift;

	# When a client trys to connect
	if(!$self->{sock}) {
		$self->{sock} = new IO::Handle;

		accept $self->{sock}, $self->{serv};

		if(!defined $self->{sock}->blocking(0)) {
			undef $self->{sock};
			return $self->err("Error with setting blocking: $!")
		}

		$self->{sock_fn} = fileno $self->{sock};

		$self->add_fh('r', $self->{sock_fn}, $self);

		$self->{sock}->autoflush(1);

		# We got 1 client, kill the server now.
		$self->disconnect_server;

		$self->debug('Recieved a client');

		return 1
	}

	$self->get_block or return $self->err;

	# We are sending the file
	if($self->{transfering}) {
		my $block;
		my $size = read $self->{FH}, $block, MAX_XFER_SIZE;

		if(!defined $size) {
			return $self->err("Could not read from file, $!")
		} elsif($size == 0) {
			return $self->err("ACK! sent $self->{sent} bytes but " .
						'EOF reached')
		}

		# Construct a header for this block
		my @header = (0, $size % 256, $size / 256);

		print {$self->{sock}} join('', map {chr} @header) . $block
					or return $self->err('Hmm');

		$self->{sent} -= $size;
		$self->debug($size, $self->{sent});

		# XXX: This could prolly just be a ->disconnect(), but I want to
		# catch what the clients do at the end of the session.
		if($self->{sent} == 0) {
			$self->del_fh('w', $self->{sock_fn});
			$self->{transfering} = 0;
			$self->{buffer} = '';		# No telling what..
		}

		# The client will send something to bring us here again.
		return 1
	}

	my $line = $self->get_line or return 1;

	my @bit = split ' ', $line;

	if($bit[0] eq $MSNCMD{version}) {
		return $self->err("$line is unsupported") if $bit[1] ne
								$MSN_FTP_VER;

		$self->send_line('version', $MSN_FTP_VER) or return $self->err
	} elsif($bit[0] eq $MSNCMD{user}) {
		# XXX: Check the user is correct
		return $self->disconnect if $bit[2] ne $self->{Auth};

		$self->{authd}++;	# Authorized

		# Send filesize in bytes
		$self->{sent} = $self->{Size};
		$self->send_line('size', $self->{Size}) or return $self->err
	} elsif($bit[0] eq $MSNCMD{quit} or $bit[0] eq $MSNCMD{out}) {
		$self->debug("Quit: $line");
		return $self->disconnect
	} elsif(!$self->{authd}) {
		return $self->err("Client attempted to do $line while not " .
					'authorized')

	# Privilaged commands now:
	} elsif($bit[0] eq $MSNCMD{transfer}) {
		$self->add_fh('w', $self->{sock_fn}, $self);
		$self->{transfering}++
	}

	push @_, $self;				# XXX: KLUGE!! :(
	redo
}}

1
